import unittest

from neatest.connection import Connection, DummyConnection, Weight, GeneRate, align_connections
from neatest.node import Node, NodeType


class TestConnection(unittest.TestCase):
    def test_eq(self):
        connection_1 = DummyConnection(Node(0, NodeType.INPUT),
                                       Node(1, NodeType.OUTPUT))
        connection_2 = DummyConnection(Node(0, NodeType.INPUT),
                                       Node(1, NodeType.OUTPUT))
        self.assertEqual(connection_1, connection_2)

    def test_hash(self):
        in_node = Node(0, NodeType.INPUT)
        out_node = Node(1, NodeType.OUTPUT)
        connection = DummyConnection(in_node, out_node)
        self.assertEqual(
            hash(str(in_node.id)+str(out_node.id)),
            hash(connection))

    def test_out_node(self):
        in_node = Node(0, NodeType.INPUT)
        out_node = Node(1, NodeType.OUTPUT)
        connection = Connection(in_node, out_node)
        self.assertEqual(out_node.inputs[0],
                         connection)

    def test_shared(self):
        weight = Weight(0.0)
        gene_rate = GeneRate(0.0)
        in_node = Node(0, NodeType.INPUT)
        out_node = Node(1, NodeType.OUTPUT)
        connection_1 = Connection(in_node, out_node, 0, gene_rate, weight)
        connection_2 = Connection(in_node, out_node, 1, gene_rate, weight)
        connection_1.weight.value = 1.0
        connection_1.dominant_gene_rate.value = 1.0
        self.assertEqual(connection_1.weight.value, connection_2.weight.value)
        self.assertEqual(connection_1.dominant_gene_rate.value,
                         connection_2.dominant_gene_rate.value)

    def test_align(self):
        in_node = Node(0, NodeType.INPUT)
        out_node = Node(1, NodeType.OUTPUT)
        a = Connection(in_node, out_node, 0)
        b = Connection(in_node, out_node, 1)
        c = Connection(in_node, out_node, 2)
        d = Connection(in_node, out_node, 3)
        e = Connection(in_node, out_node, 4)
        f = Connection(in_node, out_node, 5)
        connections_1, connections_2 = align_connections([a, b, c, f], [c, d, e])

        dummy_node = Node(0, NodeType.HIDDEN)
        dummy_connection = Connection(dummy_node, dummy_node, dummy=True)

        self.assertEqual(
            [a, b, c, dummy_connection, dummy_connection, f],
            connections_1)
        self.assertEqual(
            [dummy_connection, dummy_connection, c, d, e, dummy_connection],
            connections_2)


if __name__ == '__main__':
    unittest.main()



import functools
from itertools import groupby
from typing import Callable, List, TYPE_CHECKING
import copy
from enum import Enum

import numpy as np


def passthrough(x):
    return x


def sigmoid(x: float) -> float:
    return 1 / (1 + np.exp(-x))


def steepened_sigmoid(x: float) -> float:
    return 1 / (1 + np.exp(-4.9 * x))


def relu(x: float) -> float:
    return max(x, 0.0)


def leaky_relu(x: float, negative_slope: float = 0.01) -> float:
    return max(negative_slope*x, x)


def tanh(x: float) -> float:
    return float(np.tanh(x))


@functools.total_ordering
class NodeType(Enum):
    INPUT = 1
    BIAS = 2
    HIDDEN = 3
    OUTPUT = 4

    def __gt__(self, other) -> bool:
        return self.value > other.value


@functools.total_ordering
class Node(object):
    def __init__(self, id: int, type: NodeType,
                 activation: Callable[[float], float] = passthrough,
                 value: float = 0.0,
                 depth: float = 0.0):
        self.type = type
        self.id = id
        self.activation = activation
        self.value = value
        self.depth = depth

        if TYPE_CHECKING:
            from .connection import Connection
        self.inputs: List[Connection] = []

    def __hash__(self):
        return hash(str(self.id))

    def __eq__(self, other):
        if isinstance(other, Node):
            return self.id == other.id
        else:
            raise ValueError(f'Value type should be Node, got {type(other)}')

    def __lt__(self, other):
        if isinstance(other, Node):
            return self.id < other.id
        else:
            raise ValueError(f'Value type should be Node, got {type(other)}')

    def __add__(self, other):
        if isinstance(other, Node):
            return self.id + other.id
        else:
            return self.id + other

    def copy(self):
        return copy.deepcopy(self)

    def __str__(self):
        return f'{self.id}:{self.type}'

    def __repr__(self):
        return str(self.id)


def group_nodes(nodes: List[Node], by) -> List[List[Node]]:
    sorted_nodes = sorted(nodes, key=lambda x: getattr(x, by))
    grouped_nodes = [list(it) for k, it in groupby(
        sorted_nodes, lambda x: getattr(x, by))]
    return grouped_nodes



#!/usr/bin/env python3
from typing import List, Union, cast

import gym  # type: ignore
import numpy as np

import neatest


class Agent(neatest.Agent):
    def __init__(self):
        self.env = gym.make('CartPole-v1')

    def rollout(self, genome: neatest.Genome, render=False) -> float:
        total_reward: float = 0.0
        for i in range(10):
            observation: np.ndarray = self.env.reset()
            done: bool = False
            while not done:
                output: np.int32 = np.argmax(genome(cast(List[float], observation)))
                observation, reward, done, info = self.env.step(output)
                if render:
                    self.env.render()
                total_reward += reward
        return total_reward / 10


agent = Agent()

SEED = 123
agent.env.seed(SEED)
agent.env.action_space.seed(SEED)

neat = neatest.NEATEST(
    agent,
    neatest.Adam,
    n_networks=128,
    es_population=256,
    input_size=agent.env.observation_space.shape[0],
    output_size=agent.env.action_space.n,
    bias=True,
    dominant_gene_rate=0.5,
    dominant_gene_delta=0.05,
    elite_rate=0.05,
    sigma=0.01,
    save_checkpoint_n=50,
    optimizer_kwargs={'lr': 0.01},
    disable_connection_mutation_rate=0.3,
    seed=SEED,
    # logdir = './logs/test',
    hidden_activation=neatest.relu,
    node_mutation_rate=0.3,
    connection_mutation_rate=0.3)

neat.train(10)

print(neat.best_genome)
neat.best_genome.save('CartPole.genome')
print(agent.rollout(neat.best_genome, render=True))

agent.env.close()



import copy
from typing import List, Tuple
from itertools import chain
from dataclasses import dataclass

from .node import Node, NodeType


@dataclass
class Weight():
    value: float
    grad: float = 0.0


@dataclass
class GeneRate():
    value: float


class Connection(object):

    def __init__(self, in_node: Node, out_node: Node, innovation: int = -1,
                 dominant_gene_rate: GeneRate = GeneRate(0.0),
                 weight: Weight = Weight(1.0), dummy: bool = False):
        self.in_node = in_node
        self.out_node = out_node
        self.enabled = True
        self.dummy = dummy
        self.innovation = innovation
        self.dominant_gene_rate = dominant_gene_rate
        self.weight = weight
        if not dummy:
            self.out_node.inputs.append(self)

    def __gt__(self, other):
        return self.innovation > other.innovation

    def __lt__(self, other):
        return self.innovation < other.innovation

    def __hash__(self):
        return hash(str(self.in_node.id)+str(self.out_node.id))

    def __eq__(self, other):
        if isinstance(other, Connection):
            return ((self.in_node == other.in_node) and
                    (self.out_node == other.out_node))
        else:
            raise ValueError(f'Value type should be Connection, got {type(other)}')

    def copy(self):
        return copy.deepcopy(self)

    def __str__(self):
        string = f'{self.in_node.id} -> {self.out_node.id} '
        string = f'{string}Weight: {self.weight.value:.3f} '
        string = f'{string}Dominant Gene Rate: {self.dominant_gene_rate.value:.3f} '
        if self.dummy:
            string = f'{string}Innovation No: Dummy '
        else:
            string = f'{string}Innovation No: {self.innovation} '
        string = f'{string}Disabled: {not self.enabled} '
        return string

    def __repr__(self):
        if self.dummy:
            return 'Dummy'
        else:
            return str(self.innovation)


def align_connections(
        connections_1: List[Connection],
        connections_2: List[Connection]) -> Tuple[
            List[Connection],
            List[Connection]]:
    dummy_node = Node(0, NodeType.HIDDEN)
    dummy_connection = DummyConnection(dummy_node, dummy_node)
    end = dummy_connection
    iterators = [chain(i, [end]) for i in [sorted(connections_1), sorted(connections_2)]]
    values = [next(i) for i in iterators]
    connections_1 = []
    connections_2 = []
    while not all(v is end for v in values):
        smallest = min(v for v in values if v is not end)
        alignment = []
        for v in values:
            if v.innovation == smallest.innovation:
                alignment.append(v)
            else:
                alignment.append(dummy_connection)
        connection_1, connection_2 = alignment
        connections_1.append(connection_1)
        connections_2.append(connection_2)
        values = [next(i) if v.innovation == smallest.innovation else v
                  for i, v in zip(iterators, values)]
    return connections_1, connections_2


class DummyConnection(Connection):
    def __init__(self, in_node: Node, out_node: Node):
        super().__init__(in_node, out_node, dummy=True)



from .neatest import NEATEST, Agent
from . import connection
from . import node
from .node import Node
from .node import passthrough, sigmoid, steepened_sigmoid, relu, leaky_relu, tanh
from .genome import Genome
from .optimizers import Adam, Optimizer
from .version import VERSION

__all__ = ['NEATEST',
           'Agent',
           'connection',
           'Genome',
           'node',
           'Node',
           'passthrough',
           'sigmoid',
           'steepened_sigmoid',
           'relu',
           'leaky_relu',
           'tanh',
           'Adam',
           'Optimizer',
           'VERSION']



from typing import Union, List
from abc import ABC, abstractmethod

import numpy as np

from .connection import Weight


class Optimizer(ABC):
    def __init__(self, weights: List[Weight], **kwargs):
        self.weights = weights

    def zero_grad(self) -> None:
        for i in range(len(self.weights)):
            self.weights[i].grad = 0.0

    @abstractmethod
    def step(self) -> None:
        ...


class Adam(Optimizer):
    def __init__(self, weights: List[Weight], lr: float, beta_1: float = 0.9, beta_2:
                 float = 0.999, epsilon: float = 1e-08):

        self.lr = lr
        self.beta_1 = beta_1
        self.beta_2 = beta_2
        self.epsilon = epsilon
        self.weights = weights

        self.m: np.ndarray = np.zeros(len(weights))
        self.v: np.ndarray = np.zeros(len(weights))

        self.t: int = 0

    def step(self) -> None:
        self.t += 1
        pad_size = len(self.weights) - self.m.size
        self.m = np.concatenate([self.m, np.zeros(pad_size)])
        self.v = np.concatenate([self.v, np.zeros(pad_size)])
        gradients: np.ndarray = np.array([weight.grad for weight in self.weights])
        lr = self.lr * (np.sqrt(1 - np.power(self.beta_2, self.t)) /
                        (1 - np.power(self.beta_1, self.t)))
        self.m = self.beta_1 * self.m + (1 - self.beta_1) * gradients
        self.v = self.beta_2 * self.v + (1 - self.beta_2) * np.square(gradients)
        step = -lr * self.m / (np.sqrt(self.v) + self.epsilon)
        assert step.size == len(self.weights)
        for i in range(len(self.weights)):
            self.weights[i].value += step[i]



import unittest

from neatest.node import Node, NodeType, group_nodes


class TestNode(unittest.TestCase):
    def test_eq(self):
        a = Node(0, NodeType.INPUT)
        b = Node(0, NodeType.INPUT)
        self.assertEqual(a, b)

    def test_hash(self):
        self.assertEqual(
            hash(Node(0, NodeType.INPUT)),
            hash(str(0)))

    def test_group_by_type(self):
        a = Node(0, NodeType.INPUT)
        b = Node(1, NodeType.BIAS)
        c = Node(2, NodeType.OUTPUT)
        d = Node(3, NodeType.OUTPUT)
        e = Node(4, NodeType.HIDDEN)
        nodes = [a, b, c, d, e]
        self.assertEqual(group_nodes(nodes, by='type'),
                         [[a], [b], [e], [c, d]])

    def test_group_by_depth(self):
        a = Node(0, NodeType.INPUT, depth=0.0)
        b = Node(1, NodeType.BIAS, depth=0.0)
        c = Node(2, NodeType.OUTPUT, depth=1.0)
        d = Node(3, NodeType.OUTPUT, depth=1.0)
        e = Node(4, NodeType.HIDDEN, depth=0.5)
        nodes = [a, b, c, d, e]
        self.assertEqual(group_nodes(nodes, by='depth'),
                         [[a, b], [e], [c, d]])


if __name__ == '__main__':
    unittest.main()



from abc import ABC, abstractmethod
from typing import Union, List, Callable, Tuple, NewType, Type, Dict
from enum import Enum
import functools
import random

import numpy as np


class Version():
    major: int
    minor: int
    patch: int
    def __init__(self):
        ...

    def __eq__(self, other) -> bool:
        ...

@functools.lru_cache(maxsize=1)
def _center_function(population_size: int) -> np.ndarray:
    ...

def _compute_ranks(rewards: Union[List[float], np.ndarray]) -> np.ndarray:
    ...

def rank_transformation(rewards: Union[List[float], np.ndarray]) -> np.ndarray:
    ...

@functools.total_ordering
class NodeType(Enum):
    def __gt__(self, other: NodeType) -> bool:
        ...

def passthrough(x: float) -> float:
    ...

def sigmoid(x: float) -> float:
    ...

def steepened_sigmoid(x: float) -> float:
    ...

def relu(x: float) -> float:
    ...

def leaky_relu(x: float, negative_slope: float) -> float:
    ...

def tanh(x: float) -> float:
    ...

class Node():
    def __init__(self, id: int, type: NodeType,
                 activation: Callable[[float], float] = ...,
                 value: float = ...,
                 depth: float = ...):
        ...

    def __hash__(self) -> int:
        ...

    def __eq__(self, other: object) -> bool:
        ...

    def __lt__(self, other: object) -> bool:
        ...

    def __add__(self, other: object) -> int:
        ...

    def copy(self) -> 'Node':
        ...

    def __str__(self) -> str:
        ...

    def __repr__(self) -> str:
        ...

def group_nodes(nodes: List[Node], by: str) -> List[List[Node]]:
    ...

class Weight():
    def __init__(self, value: float, grad: float):
        ...

class GeneRate():
    def __init__(self, value: float):
        ...

class Connection(object):
    def __init__(self, in_node: Node, out_node: Node, dominant_gene_rate: GeneRate = ...,
                 weight: Weight = ..., dummy: bool = ...):
        ...

    def __gt__(self, other: 'Connection') -> bool:
        ...

    def __hash__(self) -> int:
        ...

    def __eq__(self, other: object) -> bool:
        ...

    def copy(self) -> 'Connection':
        ...

    def __str__(self) -> str:
        ...

    def __repr__(self) -> str:
        ...

def align_connections(
        connections_1: List[Connection],
        connections_2: List[Connection]) -> Tuple[
            List[Connection],
            List[Connection]]:
    ...

class DummyConnection(Connection):
    def __init__(self, in_node: Node, out_node: Node):
        ...

class Genome(object):
    input_size: int
    output_size: int
    outputs: List[List[Node]]
    version: Version

    def __init__(self, nodes: List['Node'], connections: List[Connection]):
        ...

    def __call__(self, inputs: List[float]) -> List[float]:
        ...

    @property
    def size(self) -> int:
        ...

    def deepcopy(self) -> 'Genome':
        ...

    def copy(self) -> 'Genome':
        ...

    def draw(self, node_radius: float = ...,
             vertical_distance: float = ...,
             horizontal_distance: float = ...) -> None:
        ...

    def save(self, filename: str) -> None:
        ...

    @classmethod
    def load(cls, filename: str) -> 'Genome':
        ...

    def __str__(self) -> str:
        ...

def draw_genome(genome: Genome,
                node_radius: float = ...,
                vertical_distance: float = ...,
                horizontal_distance: float = ...) -> None:
    ...

class ContextGenome(Genome):
    def __init__(self, nodes: List[Node], connections: List[Connection]) -> None:
        ...

    def copy(self) -> 'ContextGenome': #type: ignore
        ...

    def deepcopy(self) -> 'ContextGenome': #type: ignore
        ...

SortedContextGenomes = NewType('SortedContextGenomes', List[ContextGenome])

class Optimizer(ABC):
    def __init__(self, weights: List[Weight], **kwargs):
        ...

    def zero_grad(self) -> None:
        ...

    @abstractmethod
    def step(self) -> None:
        ...

class Agent(ABC):
    @abstractmethod
    def rollout(self, genome: Genome) -> float:
        ...

class Adam(Optimizer):
    m: np.ndarray
    v: np.ndarray
    t: int
    def __init__(self, weights: List[Weight], lr: float, beta_1: float = ...,
                 beta_2: float = ..., epsilon: float = ...):
        ...

    def step(self) -> None:
        ...

class NEATEST(object):
    population: List[ContextGenome]
    generation: int
    connections: Dict[Connection, int]
    best_fitness: float
    best_genome: ContextGenome
    version: Version
    n_proc: int
    random: random.Random
    np_random: np.random.RandomState
    weights: List[Weight]
    gene_rates: List[GeneRate]
    optimizer: Optimizer
    data: List[Tuple[int, str, int, float]]

    def __init__(self,
                 agent: Agent,
                 optimizer: Type[Optimizer],
                 n_networks: int,
                 es_population: int,
                 input_size: int,
                 output_size: int,
                 bias: bool,
                 node_mutation_rate: float,
                 connection_mutation_rate: float,
                 disable_connection_mutation_rate: float,
                 dominant_gene_rate: float,
                 dominant_gene_delta: float,
                 seed: int,
                 save_checkpoint_n: int = ...,
                 logdir: str = ...,
                 optimizer_kwargs: dict = ...,
                 hidden_layers: List[int] = ...,
                 elite_rate: float = ...,
                 sigma: float = ...,
                 hidden_activation: Callable[[float], float]=...,
                 output_activation: Callable[[float], float]=...):

        ...


    def register_connection(self, dummy_connection: DummyConnection):
        ...

    def add_connection_mutation(self, genome: Genome) -> None:
        ...

    def add_node_mutation(self, genome: Genome,
                          activation: Callable[[float], float]=...) -> None:
        ...

    def disable_connection_mutation(self, genome: Genome) -> None:
        ...

    def crossover(self, genome_1:Genome, genome_2:Genome) -> Genome:
        ...

    def random_genome(self) -> ContextGenome:
        ...

    def create_population(self) -> List[ContextGenome]:
        ...

    def save_logs(self) -> None:
        ...
 
    def next_generation(self, sorted_population: SortedContextGenomes) -> None:
        ...

    def calculate_grads(self, genome: ContextGenome) -> None:
        ...

    def train_genome(self, genome: ContextGenome, n_steps: int = ...):
        ...

    def reset_values(self) -> None:
        ...

    def train(self, n_steps: int) -> None:
        ...

    @staticmethod
    def sort_population(population: List[ContextGenome]) -> SortedContextGenomes:
        ...

    def get_random_genome(self) -> ContextGenome:
        ...

    def save_checkpoint(self) -> None:
        ...

    @classmethod
    def load_checkpoint(cls, filename: str) -> 'NEATEST':
        ...



import glob
import argparse
import os

import pandas as pd  # type: ignore
import numpy as np  # type: ignore
import matplotlib.pyplot as plt  # type: ignore
import seaborn as sns  # type: ignore


def plot(log_paths, smooth=1):
    sns.set(style="darkgrid", font_scale=1.5)
    estimator = getattr(np, 'mean')
    for log_path in log_paths:
        if os.path.isfile(log_path):
            logs = [log_path]
        else:
            logs = glob.glob(os.path.join(log_path, '*.csv'))
        data = [pd.read_csv(f) for f in logs]
        if smooth > 1:
            y = np.ones(smooth)
            for datum in data:
                x = np.asarray(datum['Reward'])
                z = np.ones(len(x))
                smoothed_x = np.convolve(x, y, 'same') / np.convolve(z, y, 'same')
                datum['Reward'] = smoothed_x

        data = pd.concat(data, ignore_index=True)
        data = data.astype({"Generation": int, "Algorithm": str,
                            'Seed': int, 'Reward': float})
        sns.lineplot(x="Generation", y="Reward", hue="Algorithm",
                     data=data, ci='sd', estimator=estimator)

        plt.legend(loc='best').set_draggable(True)
        plt.tight_layout(pad=0.5)
        plt.show()


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Process some integers.')
    parser.add_argument('log_paths', type=str, nargs='+', help='Log Path to Plot')
    parser.add_argument('--smooth', required=False, type=int, default=1,
                        help='Smoothing Coefficient')
    args = parser.parse_args()
    plot(args.log_paths, args.smooth)



from typing import List
import math
import os

from .connection import Connection, GeneRate, Weight
from .node import Node, NodeType, group_nodes
from .version import VERSION

import cloudpickle  # type: ignore
try:
    disable_mpi = os.environ.get('NEATEST_DISABLE_MPI')
    if disable_mpi and disable_mpi != '0':
        raise ImportError
    from mpi4py import MPI  # type: ignore
except ImportError:
    from .MPI import MPI
    MPI = MPI()


class Genome(object):
    def __init__(self, nodes: List[Node], connections: List[Connection]):
        self.connections = connections
        self.nodes = nodes
        self.version = VERSION

        grouped_nodes = group_nodes(self.nodes, 'type')
        self.input_size = len(grouped_nodes[0])
        self.output_size = len(grouped_nodes[-1])
        self.outputs = grouped_nodes[-1]

    def __call__(self, inputs: List[float]) -> List[float]:
        self.nodes.sort(key=lambda x: x.depth)
        for node in self.nodes:
            value = 0.0
            if node.type == NodeType.INPUT:
                value += inputs[node.id]
            elif node.type == NodeType.BIAS:
                continue
            for connection in node.inputs:
                if connection.enabled:
                    value += connection.in_node.value * connection.weight.value
            node.value = node.activation(value)
        return [node.value for node in self.outputs]

    def copy(self):
        connections: List[Connection] = []
        nodes: List[Node] = []
        for idx in range(len(self.connections)):
            connection = self.connections[idx]
            in_node = Node(connection.in_node.id, connection.in_node.type,
                           connection.in_node.activation,
                           depth=connection.in_node.depth)
            out_node = Node(connection.out_node.id, connection.out_node.type,
                            connection.out_node.activation,
                            depth=connection.out_node.depth)
            nodes_dict = dict(zip(nodes, range(len(nodes))))
            if in_node not in nodes_dict:
                nodes.append(in_node)
                nodes_dict[in_node] = len(nodes)-1
            if out_node not in nodes_dict:
                nodes.append(out_node)
                nodes_dict[out_node] = len(nodes)-1

            new_connection = Connection(nodes[nodes_dict[in_node]],
                                        nodes[nodes_dict[out_node]],
                                        innovation=connection.innovation,
                                        dominant_gene_rate=connection.dominant_gene_rate,
                                        weight=connection.weight)
            new_connection.enabled = connection.enabled
            connections.append(new_connection)
        new_genome = Genome(nodes, connections)
        return new_genome

    def deepcopy(self):
        connections: List[Connection] = []
        nodes: List[Node] = []
        for idx in range(len(self.connections)):
            connection = self.connections[idx]
            in_node = Node(connection.in_node.id, connection.in_node.type,
                           connection.in_node.activation,
                           depth=connection.in_node.depth)
            out_node = Node(connection.out_node.id, connection.out_node.type,
                            connection.out_node.activation,
                            depth=connection.out_node.depth)
            nodes_dict = dict(zip(nodes, range(len(nodes))))
            if in_node not in nodes_dict:
                nodes.append(in_node)
                nodes_dict[in_node] = len(nodes)-1
            if out_node not in nodes_dict:
                nodes.append(out_node)
                nodes_dict[out_node] = len(nodes)-1
            new_connection = Connection(nodes[nodes_dict[in_node]],
                                        nodes[nodes_dict[out_node]],
                                        innovation=connection.innovation,
                                        dominant_gene_rate=GeneRate(
                                            connection.dominant_gene_rate.value),
                                        weight=Weight(connection.weight.value))
            new_connection.enabled = connection.enabled
            connections.append(new_connection)
        new_genome = Genome(nodes, connections)
        return new_genome

    def draw(self, node_radius: float = 0.05,
             vertical_distance: float = 0.25,
             horizontal_distance: float = 0.25) -> None:
        draw_genome(self, node_radius, vertical_distance, horizontal_distance)

    def save(self, filename: str) -> None:
        if MPI.COMM_WORLD.rank == 0:
            with open(filename, 'wb') as output:
                cloudpickle.dump(self, output)

    @classmethod
    def load(cls, filename: str) -> 'Genome':
        print(f"\033[33;1mLoading: {filename}\033[0m")
        with open(filename, 'rb') as f:
            genome = cloudpickle.load(f)
        if genome.version != VERSION:
            print("\033[31;1mWarning: Genome version mismatch!\n"
                  f"Current Version: {VERSION.major}.{VERSION.minor}.{VERSION.patch}\n"
                  "Checkpoint Version:"
                  f" {genome.version.major}.{genome.version.minor}."
                  f"{genome.version.patch}\033[0m")
        return genome

    def __str__(self):
        string = ''
        string = f'{string}NODES:\n'
        for node in self.nodes:
            string = f'{string}{node}\n'
        string = f'{string}\n\nCONNECTIONS:\n'
        for connection in self.connections:
            string = f'{string}{connection}\n'
        return string


def draw_genome(genome: Genome,
                node_radius: float = 0.05,
                vertical_distance: float = 0.25,
                horizontal_distance: float = 0.25) -> None:
    '''Draw the genome to a matplotlib figure but do not show it.'''
    import matplotlib.pyplot as plt  # type: ignore
    import matplotlib.patches as patches  # type: ignore
    plt.gcf().canvas.set_window_title('float')

    positions = {}
    node_groups = group_nodes(genome.nodes, 'depth')
    for group_idx, nodes in enumerate(node_groups):
        y_position = -vertical_distance * (len(nodes)-1)/2
        for i, node in enumerate(nodes):
            positions[f'{node.id}'] = (group_idx * horizontal_distance,
                                       y_position + i*vertical_distance)
            circle = plt.Circle(positions[f'{node.id}'],
                                node_radius, color='r', fill=False)
            plt.gcf().gca().text(*positions[f'{node.id}'], node.id,
                                 horizontalalignment='center',
                                 verticalalignment='center',
                                 fontsize=10.0)
            plt.gcf().gca().add_artist(circle)

    for connection in genome.connections:
        if connection.enabled:
            node1_x = positions[f'{connection.in_node.id}'][0]
            node2_x = positions[f'{connection.out_node.id}'][0]
            node1_y = positions[f'{connection.in_node.id}'][1]
            node2_y = positions[f'{connection.out_node.id}'][1]
            angle = math.atan2(node2_x - node1_x, node2_y - node1_y)
            x_adjustment = node_radius * math.sin(angle)
            y_adjustment = node_radius * math.cos(angle)
            arrow = patches.FancyArrowPatch(
                (node1_x + x_adjustment,
                 node1_y + y_adjustment),
                (node2_x - x_adjustment,
                 node2_y - y_adjustment),
                arrowstyle="Simple,tail_width=0.5,head_width=3,head_length=5",
                color="k", antialiased=True)
            plt.gcf().gca().add_patch(arrow)
    plt.axis('scaled')



import unittest

from neatest.connection import Connection, Weight, GeneRate
from neatest.node import Node, NodeType
from neatest.genome import Genome


class TestGenome(unittest.TestCase):
    def setUp(self):
        na = Node(0, NodeType.INPUT, depth=0.0)
        nb = Node(1, NodeType.INPUT, depth=0.0)
        nc = Node(2, NodeType.BIAS, value=1.0, depth=0.0)
        nd = Node(3, NodeType.OUTPUT, depth=1.0)
        ne = Node(4, NodeType.OUTPUT, depth=1.0)
        nf = Node(5, NodeType.HIDDEN, depth=0.5)
        ca = Connection(na, nd, 0, GeneRate(0.0), Weight(1.0))
        cb = Connection(na, ne, 1, GeneRate(0.0), Weight(1.0))
        cc = Connection(nb, nd, 2, GeneRate(0.0), Weight(1.0))
        cd = Connection(nb, ne, 3, GeneRate(0.0), Weight(1.0))
        ce = Connection(nc, nd, 4, GeneRate(0.0), Weight(1.0))
        cf = Connection(nc, ne, 5, GeneRate(0.0), Weight(1.0))
        cg = Connection(ne, nf, 6, GeneRate(0.0), Weight(1.0))
        ch = Connection(nf, ne, 7, GeneRate(0.0), Weight(1.0))
        self.genome = Genome([na, nb, nc, nd, ne],
                             [ca, cb, cc, cd, ce, cf])
        self.recursive_genome = Genome([na, nb, nc, nd, ne, nf],
                                       [ca, cb, cc, cd, ce, cf, cg, ch])

    def test_init(self):
        self.assertEqual(self.genome.input_size, 2)
        self.assertEqual(self.genome.output_size, 2)

    def test_call(self):
        self.assertEqual(self.genome([1.0, 1.0]), [3.0, 3.0])
        self.assertEqual(self.genome([1.0, 1.0]), [3.0, 3.0])
        self.assertEqual(self.genome([1.0, 1.0]), [3.0, 3.0])
        self.assertEqual(self.genome([1.0, 1.0]), [3.0, 3.0])

    def test_recursive_call(self):
        self.assertEqual(self.recursive_genome([1.0, 1.0]), [3.0, 3.0])
        self.assertEqual(self.recursive_genome([1.0, 1.0]), [3.0, 6.0])
        self.assertEqual(self.recursive_genome([1.0, 1.0]), [3.0, 9.0])
        self.assertEqual(self.recursive_genome([1.0, 1.0]), [3.0, 12.0])

    def test_copy(self):
        genome = self.genome.copy()
        genome.connections[0].weight.value = 2.0
        genome.connections[0].dominant_gene_rate.value = 2.0
        self.assertEqual(self.genome.connections[0].weight.value, 2.0)
        self.assertEqual(self.genome.connections[0].dominant_gene_rate.value, 2.0)
        self.assertIs(self.genome.connections[0].weight, genome.connections[0].weight)
        self.assertIs(self.genome.connections[0].dominant_gene_rate,
                      genome.connections[0].dominant_gene_rate)

    def test_deepcopy(self):
        genome = self.genome.deepcopy()
        genome.connections[1].weight.value = 2.0
        genome.connections[1].dominant_gene_rate.value = 2.0
        self.assertEqual(self.genome.connections[1].weight.value, 1.0)
        self.assertEqual(self.genome.connections[1].dominant_gene_rate.value, 0.0)
        self.assertIsNot(self.genome.connections[1].weight, genome.connections[1].weight)
        self.assertIsNot(self.genome.connections[1].dominant_gene_rate,
                         genome.connections[1].dominant_gene_rate)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
import random
from typing import Union, List, Callable, Tuple, NewType, Dict, Type, Sequence
import itertools
import functools
from abc import ABC, abstractmethod
import operator
import os
import sys
import time
import pathlib
import inspect

try:
    disable_mpi = os.environ.get('NEATEST_DISABLE_MPI')
    if disable_mpi and disable_mpi != '0':
        raise ImportError
    from mpi4py import MPI  # type: ignore
except ImportError:
    from .MPI import MPI
    MPI = MPI()
import numpy as np
import cloudpickle  # type: ignore

from .genome import Genome
from .node import Node, NodeType, passthrough
from .connection import Connection, Weight, GeneRate, DummyConnection, align_connections
from .optimizers import Optimizer
from .version import VERSION



@functools.lru_cache(maxsize=1)
def _center_function(population_size: int) -> np.ndarray:
    centers = np.arange(0, population_size)
    centers = centers / (population_size - 1)
    centers -= 0.5
    return centers


def _compute_ranks(rewards: Union[List[float], np.ndarray]) -> np.ndarray:
    rewards = np.array(rewards)
    ranks = np.empty(rewards.size, dtype=int)
    ranks[rewards.argsort()] = np.arange(rewards.size)
    return ranks


def rank_transformation(rewards: Union[List[float], np.ndarray]) -> np.ndarray:
    ranks = _compute_ranks(rewards)
    values = _center_function(len(rewards))
    return values[ranks]  # type: ignore


class ContextGenome(Genome):
    '''Genome class that holds data which depends on the context.'''
    def __init__(self, nodes: List[Node], connections: List[Connection]):
        self.fitness: float = 0.0
        self.generation: int = 1
        super().__init__(nodes, connections)

    def copy(self) -> 'ContextGenome':  # type: ignore
        new_genome = super(ContextGenome, self).copy()
        return ContextGenome(new_genome.nodes, new_genome.connections)

    def deepcopy(self) -> 'ContextGenome':  # type: ignore
        new_genome = super(ContextGenome, self).deepcopy()
        return ContextGenome(new_genome.nodes, new_genome.connections)


SortedContextGenomes = NewType('SortedContextGenomes', List[ContextGenome])


class Agent(ABC):
    @abstractmethod
    def rollout(self, genome: Genome) -> float:
        ...


class NEATEST(object):
    def __init__(self,
                 agent: Agent,
                 optimizer: Type[Optimizer],
                 n_networks: int,
                 es_population: int,
                 input_size: int,
                 output_size: int,
                 bias: bool,
                 node_mutation_rate: float,
                 connection_mutation_rate: float,
                 disable_connection_mutation_rate: float,
                 dominant_gene_rate: float,
                 dominant_gene_delta: float,
                 seed: int,
                 save_checkpoint_n: int = 50,
                 logdir: str = None,
                 optimizer_kwargs: dict = {},
                 hidden_layers: List[int] = [],
                 elite_rate: float = 0.0,
                 sigma: float = 0.01,
                 hidden_activation: Callable[[float], float] = passthrough,
                 output_activation: Callable[[float], float] = passthrough):

        comm = MPI.COMM_WORLD
        self.version = VERSION
        n_proc = comm.Get_size()
        self.n_proc = n_proc
        # assert not n_networks % n_proc
        # assert not es_population % n_proc
        self.seed = seed
        self.random = random.Random(seed)
        self.np_random = np.random.RandomState(seed)
        self.logdir = logdir
        self.agent = agent
        self.es_population = es_population
        self.n_networks = n_networks
        self.input_size = input_size
        self.output_size = output_size
        self.hidden_layers = hidden_layers
        self.bias = bias
        self.sigma = sigma
        self.save_checkpoint_n = save_checkpoint_n
        self.elite_rate = elite_rate
        self.disable_connection_mutation_rate = disable_connection_mutation_rate
        self.node_mutation_rate = node_mutation_rate
        self.connection_mutation_rate = connection_mutation_rate
        self.dominant_gene_rate = dominant_gene_rate
        self.dominant_gene_delta = dominant_gene_delta
        self.hidden_activation = hidden_activation
        self.output_activation = output_activation

        self.weights: List[Weight] = []
        self.gene_rates: List[GeneRate] = []
        self.connections: Dict[Connection, int] = {}
        self.node_id_generator = itertools.count(0, 1)
        self.connection_id_generator = itertools.count(0, 1)

        self.optimizer: Optimizer = optimizer(self.weights, **optimizer_kwargs)

        self.generation: int = 1
        self.best_fitness: float = -float('inf')
        self.best_genome: ContextGenome
        self.population: List[ContextGenome]

        self.data: List[Tuple[int, str, int, float]] = []

        if not comm.rank == 0:
            f = open(os.devnull, 'w')
            sys.stdout = f
        else:
            if self.logdir:
                from pathlib import Path
                Path(self.logdir).mkdir(parents=True, exist_ok=True)


        self.population = self.create_population()

    def random_genome(self) -> ContextGenome:
        '''Create fc neural network with random weights.'''
        layers = []
        connections: List[Connection] = []

        # Input Nodes
        input_nodes = [Node(next(self.node_id_generator), NodeType.INPUT, depth=0.0)
                       for _ in range(self.input_size)]
        if self.bias:
            input_nodes.append(Node(next(self.node_id_generator), NodeType.BIAS,
                                    value=1.0, depth=0.0))
        layers.append(input_nodes)

        # Hidden Nodes
        for idx, hidden_layer in enumerate(self.hidden_layers):
            depth = 1 / (len(self.hidden_layers) + 1) * (idx + 1)
            hidden_nodes = [Node(next(self.node_id_generator),
                                 NodeType.HIDDEN,
                                 self.hidden_activation,
                                 depth=depth)
                            for _ in range(hidden_layer)]
            if self.bias:
                hidden_nodes.append(
                    Node(next(self.node_id_generator),
                         NodeType.BIAS, value=1.0, depth=depth))
            layers.append(hidden_nodes)

        # Output Nodes
        output_nodes = [Node(next(self.node_id_generator), NodeType.OUTPUT,
                             self.output_activation, depth=1.0)
                        for _ in range(self.output_size)]
        layers.append(output_nodes)

        for i in range(1, len(layers)):
            for j in range(len(layers[i-1])):
                input_node = layers[i-1][j]
                for k in range(len(layers[i])):
                    output_node = layers[i][k]
                    dummy_connection = DummyConnection(input_node, output_node)
                    innovation, weight, dominant_gene_rate = self.register_connection(
                        dummy_connection)

                    connections += [Connection(
                        in_node=input_node, out_node=output_node,
                        innovation=innovation,
                        dominant_gene_rate=dominant_gene_rate,
                        weight=weight)]

        nodes: List[Node] = functools.reduce(operator.iconcat, layers, [])
        return ContextGenome(nodes, connections)

    def create_population(self) -> List[ContextGenome]:
        population: List[ContextGenome] = [self.random_genome()]
        for _ in range(self.n_networks - 1):
            population.append(population[0].copy())
        return population

    def register_connection(self, dummy_connection: DummyConnection):
        if dummy_connection in self.connections:
            innovation = self.connections[dummy_connection]
            weight = self.weights[innovation]
            dominant_gene_rate = self.gene_rates[innovation]
        else:
            innovation = next(self.connection_id_generator)
            weight = Weight(self.random.gauss(0.0, self.sigma))
            dominant_gene_rate = GeneRate(self.dominant_gene_rate)
            self.weights.append(weight)
            self.gene_rates.append(dominant_gene_rate)
            self.connections[dummy_connection] = innovation
        return innovation, weight, dominant_gene_rate

    def add_connection_mutation(self, genome: Genome) -> None:
        '''Create new connection between two random non-connected nodes.'''
        def _add_connection_mutation(depth=0):
            if depth > 20:
                return
            in_idx = self.random.randint(0, len(genome.nodes) - 1)
            in_node = genome.nodes[in_idx]
            out_idx = self.random.randint(0, len(genome.nodes) - 1)
            out_node = genome.nodes[out_idx]
            dummy_connection = DummyConnection(in_node, out_node)
            try:
                index = genome.connections.index(dummy_connection)
                if not genome.connections[index].enabled:
                    if (self.random.random() <=
                            genome.connections[index].dominant_gene_rate.value):
                        genome.connections[index].enabled = True
                        return
                else:
                    _add_connection_mutation(depth+1)
                    return
            except ValueError:
                pass

            if (out_node.type == NodeType.BIAS or
                    out_node.type == NodeType.INPUT or
                    in_node.type == NodeType.OUTPUT):
                _add_connection_mutation(depth+1)
                return
            innovation, weight, dominant_gene_rate = self.register_connection(
                dummy_connection)

            connection = Connection(in_node=in_node, out_node=out_node,
                                    innovation=innovation,
                                    dominant_gene_rate=dominant_gene_rate,
                                    weight=weight)
            genome.connections.append(connection)
        _add_connection_mutation()

    def add_node_mutation(self, genome: Genome,
                          activation: Callable[[float], float] = lambda x: x) -> None:
        '''Add a node to a random connection and split the connection.'''
        idx = self.random.randint(0, len(genome.connections)-1)
        genome.connections[idx].enabled = False

        # new_node = Node(next(self.node_id_generator), NodeType.HIDDEN, activation)
        new_node = Node(max(genome.nodes) + 1, NodeType.HIDDEN, activation)

        first_weight = Weight(1.0)
        first_gene_rate = GeneRate(self.dominant_gene_rate)
        first_innovation = next(self.connection_id_generator)
        first_connection = Connection(in_node=genome.connections[idx].in_node,
                                      out_node=new_node,
                                      innovation=first_innovation,
                                      dominant_gene_rate=first_gene_rate,
                                      weight=first_weight)
        self.weights.append(first_weight)
        self.gene_rates.append(first_gene_rate)
        self.connections[first_connection] = first_innovation

        second_weight = Weight(genome.connections[idx].weight.value)
        second_gene_rate = GeneRate(self.dominant_gene_rate)
        second_innovation = next(self.connection_id_generator)
        second_connection = Connection(in_node=new_node,
                                       out_node=genome.connections[idx].out_node,
                                       innovation=second_innovation,
                                       dominant_gene_rate=second_gene_rate,
                                       weight=second_weight)
        self.weights.append(second_weight)
        self.gene_rates.append(second_gene_rate)
        self.connections[second_connection] = second_innovation

        genome.connections.append(first_connection)
        genome.connections.append(second_connection)
        new_node.depth = (first_connection.in_node.depth +
                          second_connection.out_node.depth) / 2
        genome.nodes.append(new_node)

    def disable_connection_mutation(self, genome: Genome) -> None:
        def _disable_connection_mutation(depth=0):
            if depth > 20:
                return
            idx = self.random.randint(0, len(genome.connections)-1)
            if (genome.connections[idx].out_node.type == NodeType.OUTPUT or
                    genome.connections[idx].in_node.type == NodeType.INPUT or
                    genome.connections[idx].in_node.type == NodeType.BIAS):
                _disable_connection_mutation(depth + 1)
                return
            else:
                if not genome.connections[idx].enabled:
                    _disable_connection_mutation(depth + 1)
                    return
                else:
                    genome.connections[idx].enabled = False
                    return
        _disable_connection_mutation()

    def next_generation(self, sorted_population: SortedContextGenomes):
        population: List[ContextGenome]
        if self.elite_rate > 0.0:
            population = sorted_population[0:int(self.n_networks * self.elite_rate)]
        else:
            population = []

        self.generation += 1
        while len(population) < self.n_networks:
            genome_1 = self.get_random_genome()
            genome_2 = self.get_random_genome()
            new_genome = self.crossover(genome_1, genome_2)
            if self.random.random() < self.disable_connection_mutation_rate:
                self.disable_connection_mutation(new_genome)
            if self.random.random() < self.node_mutation_rate:
                self.add_node_mutation(
                    new_genome,
                    activation=self.hidden_activation)
            if self.random.random() < self.connection_mutation_rate:
                self.add_connection_mutation(
                    new_genome)
            new_genome.generation = self.generation
            population.append(new_genome)

        self.population = population

    def calculate_grads(self, genome: ContextGenome):
        comm = MPI.COMM_WORLD
        cp_genome: ContextGenome = genome.deepcopy()  # type: ignore
        for i in reversed(range(len(cp_genome.connections))):
            if not cp_genome.connections[i].enabled:
                del cp_genome.connections[i]
        weights: List[float] = [connection.weight.value
                                for connection in cp_genome.connections]
        weights_array: np.ndarray = np.array(weights)
        epsilon: np.ndarray = self.np_random.normal(
            0.0, self.sigma, (self.es_population//2, len(weights)))
        population_weights: np.ndarray = np.concatenate([weights_array + epsilon,
                                                    weights_array - epsilon])

        n_jobs = self.es_population // comm.Get_size()
        rewards: List[float] = []
        rewards_array: np.ndarray = np.zeros(self.es_population, dtype='d')
        for i in range(comm.rank*n_jobs, n_jobs * (comm.rank + 1)):
            for j, connection in enumerate(cp_genome.connections):
                connection.weight = Weight(population_weights[i, j])  # type: ignore
            rewards.append(self.agent.rollout(cp_genome))
        comm.Allgatherv([np.array(rewards, dtype=np.float64), MPI.DOUBLE],
                        [rewards_array, MPI.DOUBLE])
        ranked_rewards: np.ndarray = rank_transformation(rewards_array)
        epsilon = np.concatenate([epsilon, -epsilon])
        grads: np.ndarray = (np.dot(ranked_rewards, epsilon) / (self.es_population * self.sigma))
        grads = np.clip(grads, -1.0, 1.0)

        for gene_rate in self.gene_rates:
            gene_rate.value -= self.dominant_gene_delta

        idx = 0
        for connection in genome.connections:
            if connection.enabled:
                connection.weight.grad = -grads[idx]  # type: ignore
                connection.dominant_gene_rate.value += 2 * self.dominant_gene_delta
                idx += 1

        for gene_rate in self.gene_rates:
            gene_rate.value = min(0.6, max(0.4, gene_rate.value))

    def train(self, n_steps: int) -> None:
        comm = MPI.COMM_WORLD
        n_jobs = self.n_networks // comm.Get_size()
        for step in range(n_steps):
            rewards = []
            for genome in self.population[
                    comm.rank*n_jobs: n_jobs * (comm.rank + 1)]:
                reward = self.agent.rollout(genome)
                rewards.append(reward)
            rewards = functools.reduce(
                operator.iconcat, comm.allgather(rewards), [])

            for idx, reward in enumerate(rewards):
                self.population[idx].fitness = reward

            self.train_genome(self.get_random_genome())

            sorted_population: SortedContextGenomes = self.sort_population(
                self.population)

            reward = self.agent.rollout(sorted_population[0])
            self.data.append((int(self.generation), 'NEATEST', self.seed, reward))
            if reward >= self.best_fitness:
                self.best_fitness = reward
                self.best_genome = sorted_population[0].deepcopy()

            print(f'Generation: {self.generation}')
            print(f'Rollout Reward: {reward}')
            print(f'Max Reward Session: {self.best_fitness}')
            print(f'Max Reward in Population: {max(rewards)}')

            if self.generation and not self.generation % self.save_checkpoint_n:
                self.save_checkpoint()

            self.next_generation(sorted_population)

        if self.logdir:
            if MPI.COMM_WORLD.rank == 0:
                self.save_logs()

    def train_genome(self, genome: ContextGenome, n_steps: int = 1):
        for _ in range(n_steps):
            self.optimizer.zero_grad()
            self.calculate_grads(genome)
            self.optimizer.step()

    def crossover(self, genome_1: Genome, genome_2: Genome) -> ContextGenome:
        '''Crossover two genomes by aligning their innovation numbers.'''
        connections: List[Connection] = []
        nodes: List[Node] = []
        connections_1, connections_2 = align_connections(
            genome_1.connections, genome_2.connections)

        for idx in range(len(connections_1)):
            connection_1 = connections_1[idx]
            connection_2 = connections_2[idx]
            enabled: bool
            connection: Connection
            if connection_1.dummy or connection_2.dummy:
                if connection_1.dummy:
                    connection = connection_2
                else:
                    connection = connection_1
                if connection.enabled:
                    if self.random.random() <= connection.dominant_gene_rate.value:
                        enabled = True
                    else:
                        enabled = False
                else:
                    continue
            else:
                connection = connection_1
                if connection_1.enabled and connection_2.enabled:
                    enabled = True
                elif connection_1.enabled ^ connection_2.enabled:
                    enabled = (self.random.random() <=
                               connection.dominant_gene_rate.value)
                else:
                    if self.random.random() <= connection.dominant_gene_rate.value:
                        enabled = False
                    else:
                        continue

            in_node = Node(connection.in_node.id, connection.in_node.type,
                           connection.in_node.activation,
                           depth=connection.in_node.depth)
            out_node = Node(connection.out_node.id, connection.out_node.type,
                            connection.out_node.activation,
                            depth=connection.out_node.depth)

            nodes_dict = dict(zip(nodes, range(len(nodes))))
            if in_node not in nodes_dict:
                nodes.append(in_node)
                nodes_dict[in_node] = len(nodes)-1
            if out_node not in nodes_dict:
                nodes.append(out_node)
                nodes_dict[out_node] = len(nodes)-1
            connection = Connection(nodes[nodes_dict[in_node]],
                                    nodes[nodes_dict[out_node]],
                                    innovation=connection.innovation,
                                    weight=connection.weight,
                                    dominant_gene_rate=connection.dominant_gene_rate)
            connection.enabled = enabled
            connections.append(connection)
        new_genome = ContextGenome(nodes, connections)
        return new_genome

    def save_logs(self):
        import pandas as pd  # type: ignore
        data = pd.DataFrame(
            self.data, columns=['Generation', 'Algorithm', 'Seed', 'Reward'])
        data = data.astype(
            {"Generation": int, "Algorithm": str, 'Seed': int, 'Reward': float})
        file = os.path.join(self.logdir, f'{time.strftime("%Y%m%d-%H%M%S")}.csv')
        data.to_csv(file, index=False)

    @staticmethod
    def sort_population(population: List[ContextGenome]) -> SortedContextGenomes:
        return SortedContextGenomes(sorted(population, key=lambda x: x.fitness,
                                           reverse=True))

    def get_random_genome(self) -> ContextGenome:
        """Return random genome from a sorted population."""
        rewards: np.ndarray = np.array([genome.fitness for genome in self.population])
        eps = np.finfo(float).eps
        normalized_rewards: np.ndarray = rewards - rewards.min() + eps
        probabilities = normalized_rewards / np.sum(normalized_rewards)
        return self.np_random.choice(self.population, p=probabilities)

    def save_checkpoint(self) -> None:
        if MPI.COMM_WORLD.rank == 0:
            frame = inspect.stack()[2]
            module = inspect.getmodule(frame[0])
            if module is not None:
                folder = pathlib.Path(
                    f"{os.path.join(os.path.dirname(module.__file__), 'checkpoints')}")
                folder.mkdir(parents=True, exist_ok=True)
                filename = f'{int(time.time())}.checkpoint'
                save_path = os.path.abspath(os.path.join(folder, filename))
                print(f"\033[33;1mCheckpoint: {save_path}\033[0m")
                with open(save_path, 'wb') as output:
                    cloudpickle.dump(self, output)

    @classmethod
    def load_checkpoint(cls, filename: str) -> 'NEATEST':
        comm = MPI.COMM_WORLD
        if not comm.rank == 0:
            f = open(os.devnull, 'w')
            sys.stdout = f
        print(f"\033[33;1mLoading: {filename}\033[0m")
        with open(filename, 'rb') as checkpoint:
            neatest = cloudpickle.load(checkpoint)
        if comm.rank == 0:
            if neatest.logdir:
                from pathlib import Path
                Path(neatest.logdir).mkdir(parents=True, exist_ok=True)
        n_proc = comm.Get_size()
        # assert not neatest.n_networks % n_proc
        # assert not neatest.es_population % n_proc
        if neatest.n_proc != n_proc:
            print("\033[31;1mWarning: Number of process mismatch\033[0m")
        if neatest.version != VERSION:
            print("\033[31;1mWarning: Checkpoint version mismatch!\n"
                  f"Current Version: {VERSION.major}.{VERSION.minor}.{VERSION.patch}\n"
                  "Checkpoint Version:"
                  f" {neatest.version.major}.{neatest.version.minor}."
                  f"{neatest.version.patch}\n\033[0m")
        return neatest
